@@ -10,18 +10,25 @@ except ImportError:
     pass # pwd,grp modules not available on Windows
 
 from visidata import Column, Sheet, LazyComputeRow, asynccache, BaseSheet, vd
-from visidata import Path, ENTER, date, asyncthread, FileExistsError, VisiData
+from visidata import Path, ENTER, asyncthread, VisiData
 from visidata import modtime, filesize, vstat, Progress, TextSheet
+from visidata.type_date import date
 
 
-vd.option('dir_recurse', False, 'walk source path recursively on DirSheet')
+vd.option('dir_depth', 0, 'folder recursion depth on DirSheet')
 vd.option('dir_hidden', False, 'load hidden files on DirSheet')
 
 
+@VisiData.api
+def guess_dir(vd, p):
+    if p.is_dir():
+        return dict(filetype='dir')
+
+
 @VisiData.lazy_property
 def currentDirSheet(p):
     'Support opening the current DirSheet from the vdmenu'
-    return DirSheet('.', source=Path('.'))
+    return DirSheet(Path('.').absolute().name, source=Path('.'))
 
 @asyncthread
 def exec_shell(*args):
@@ -34,11 +41,11 @@ def exec_shell(*args):
 
 @VisiData.api
 def open_dir(vd, p):
-    return DirSheet(p.name, source=p)
+    return DirSheet(p.base_stem, source=p)
 
 @VisiData.api
 def open_fdir(vd, p):
-    return FileListSheet(p.name, source=p)
+    return FileListSheet(p.base_stem, source=p)
 
 @VisiData.api
 def addShellColumns(vd, cmd, sheet):
@@ -75,6 +82,21 @@ class ColumnShell(Column):
 
 class DirSheet(Sheet):
     'Sheet displaying directory, using ENTER to open a particular file.  Edited fields are applied to the filesystem.'
+    guide = '''
+        # Directory Sheet
+        This is a list of files in the {sheet.displaySource} folder.
+
+        - {help.commands.open_row_file}
+        - {help.commands.open_rows}
+        - (`open-dir-parent`) to open parent directory
+        - {help.commands.sysopen_row}
+
+        ## Options (must reload to take effect)
+
+        - {help.options.dir_depth}
+            - [CLI] `-r` to include all files in all subfolders
+        - {help.options.dir_hidden}
+    '''
     rowtype = 'files' # rowdef: Path
     defer = True
     columns = [
@@ -82,7 +104,7 @@ class DirSheet(Sheet):
             getter=lambda col,row: str(row.parent) if str(row.parent) == '.' else str(row.parent) + '/',
             setter=lambda col,row,val: col.sheet.moveFile(row, val)),
         Column('filename',
-            getter=lambda col,row: row.name + ''.join(row.suffixes),
+            getter=lambda col,row: row._path.name,
             setter=lambda col,row,val: col.sheet.renameFile(row, val)),
         Column('abspath', width=0, type=str,
             getter=lambda col,row: row,
@@ -144,11 +166,14 @@ class DirSheet(Sheet):
 
     def removeFile(self, path):
         if path.is_dir():
-            os.rmdir(path)
+            if self.options.safety_first:
+                os.rmdir(path)
+            else:
+                shutil.rmtree(path)  #1965
         else:
             path.unlink()
 
-    def deleteSourceRow(self, r):
+    def commitDeleteRow(self, r):
         self.removeFile(r)
 
     def newRow(self):
@@ -157,7 +182,7 @@ class DirSheet(Sheet):
     def iterload(self):
         hidden_files = self.options.dir_hidden
 
-        def _walkfiles(p):
+        def _walkfiles(p, dir_depth:int=0):
             basepath = str(p)
             for folder, subdirs, files in os.walk(basepath):
                 subfolder = folder[len(basepath)+1:]
@@ -165,24 +190,23 @@ class DirSheet(Sheet):
                 if subfolder in ['.', '..']: continue
 
                 fpath = Path(folder)
-                yield fpath
+                if str(fpath) != str(p):
+                    yield fpath
 
                 for fn in files:
                     yield fpath/fn
 
-        def _listfiles(p):
-            basepath = str(p)
-            for fn in os.listdir(basepath):
-                yield p/fn
-
+                if dir_depth < len(fpath.parents)-len(p.parents)+1:
+                    for d in subdirs:
+                        yield fpath/d
+                    subdirs.clear()
 
         basepath = str(self.source)
 
         folders = set()
-        f = _walkfiles if self.options.dir_recurse else _listfiles
 
-        for p in f(self.source):
-            if not hidden_files and p.name.startswith('.'):
+        for p in _walkfiles(self.source, self.options.dir_depth):
+            if not hidden_files and str(p).startswith('.') and not str(p).startswith('..'):
                 continue
 
             yield p
@@ -203,11 +227,14 @@ class DirSheet(Sheet):
         self._deferredDels.clear()
         self.reload()
 
+    def getDefaultSaveName(sheet):
+        return sheet.name + '.' + sheet.options.save_filetype
+
 
 class FileListSheet(DirSheet):
     _ordering = []
     def iterload(self):
-        for fn in self.source.open_text():
+        for fn in self.source.open():
             yield Path(fn.rstrip())
 
 
@@ -218,9 +245,10 @@ def inputShell(vd):
         vd.warning('no $column in command')
     return cmd
 
+DirSheet.addCommand('`', 'open-dir-parent', 'vd.push(openSource(source.parent if source.resolve()!=Path(".").resolve() else os.path.dirname(source.resolve())))', 'open parent directory')  #1801
 BaseSheet.addCommand('', 'open-dir-current', 'vd.push(vd.currentDirSheet)', 'open Directory Sheet: browse properties of files in current directory')
 
-Sheet.addCommand('z;', 'addcol-sh', 'cmd=inputShell(); addShellColumns(cmd, sheet)', 'create new column from bash expression, with $columnNames as variables')
+Sheet.addCommand('z;', 'addcol-shell', 'cmd=inputShell(); addShellColumns(cmd, sheet)', 'create new column from bash expression, with $columnNames as variables')
 
 DirSheet.addCommand(ENTER, 'open-row-file', 'vd.push(openSource(cursorRow or fail("no row"), filetype="dir" if cursorRow.is_dir() else LazyComputeRow(sheet, cursorRow).ext))', 'open current file as a new sheet')
 DirSheet.addCommand('g'+ENTER, 'open-rows', 'for r in selectedRows: vd.push(openSource(r))', 'open selected files as new sheets')
@@ -241,7 +269,7 @@ def copy_files(sheet, paths, dest):
         try:
             destpath = destdir/str(srcpath._path.name)
             if srcpath.is_dir():
-                shutil.copy_tree(srcpath, destpath)
+                shutil.copytree(srcpath, destpath)
             else:
                 shutil.copyfile(srcpath, destpath)
         except Exception as e:
@@ -251,3 +279,7 @@ def copy_files(sheet, paths, dest):
 vd.addGlobals({
     'DirSheet': DirSheet
 })
+
+vd.addMenuItems('''
+    Column > Add column > shell > addcol-shell
+''')
